from collections.abc import (
    AsyncIterable,
    AsyncIterator,
    Callable,
    Coroutine,
    Hashable,
    Iterable,
    Iterator,
)
from dataclasses import dataclass
from typing import Any, Literal, Protocol

type KeyFn[**P] = Callable[P, Hashable]

type Coro[T] = Coroutine[Any, Any, T]
type AnyIterable[T] = AsyncIterable[T] | Iterable[T]
type AnyIterator[T] = AsyncIterator[T] | Iterator[T]

type Get[T] = Callable[[], T]
type Callback[T] = Callable[[T], object]

type CachePolicy = Literal['lru', 'mru'] | None


@dataclass(frozen=True, slots=True)
class Some[T]:
    x: T


class Decorator(Protocol):
    def __call__[**P, R](self, fn: Callable[P, R], /) -> Callable[P, R]: ...


class PsDecorator[**P](Protocol):
    def __call__[R](self, fn: Callable[P, R], /) -> Callable[P, R]: ...



"""argparse and dataclasses, married.

Example:
```
@dataclass
class Args:
    a: int
    b: str = 'hello'

args, parser = parse_args(Args)
```
Or with plain function:
```
@parse_args
def main(name: str = 'user'):
    print(f'Hello {name}')
```

Reasons not to use alternatives:
- [simple-parsing](https://github.com/lebrice/SimpleParsing):
  - Has underscores (`--like_this`) instead of dashes (`--like-this`)
  - Erases type on parser result, thus making typed prototype useless
    (what is the point of using dataclasses if it is passed
     through function returning typing.Any/argparse.Namespace?)

- [datargs](https://github.com/roee30/datargs):
  - No nesting support
  - No function's support

- [pydantic](https://github.com/samuelcolvin/pydantic):
  - supports CLI via BaseSettings and environment variables parsing
  - no nesting, as requires mixing only via multiple inheritance

- [typer](https://github.com/tiangolo/typer):
  - No support on dataclasses
    (https://github.com/tiangolo/typer/issues/154).
  - No fine way to extract parsed options without invoking because of
    decorator/callback based implementation. Thus enforces wrapping of the
    whole app into `typer.run`.
    (https://github.com/tiangolo/typer/issues/197).
"""

__all__ = ['arg', 'parse_args']

import argparse
import importlib
import sys
import types
from argparse import ArgumentParser, BooleanOptionalAction, _ArgumentGroup
from collections.abc import Callable, Iterable, Iterator, Sequence
from dataclasses import MISSING, Field, dataclass, field, fields, is_dataclass
from inspect import getmodule, signature, stack
from typing import (
    Any,
    Literal,
    Required,
    TypedDict,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

from typing_inspection.introspection import (
    UNKNOWN,
    AnnotationSource,
    inspect_annotation,
)

type _Node = str | tuple[str, type, list['_Node']]


@dataclass(kw_only=True)
class Meta:
    help: str = ''
    flag: str | None = None


@dataclass(kw_only=True)
class _Meta(Meta):
    name: str


def arg(
    default=MISSING,
    /,
    *,
    flag=None,
    factory=MISSING,
    init=True,
    repr=True,  # noqa: A002
    hash=None,  # noqa: A002
    help=None,  # noqa: A002
    compare=True,
    metadata=None,
) -> Field:
    """Annotate dataclass.field with extra metadata (like help)."""
    metadata = metadata or {}
    for k, v in {'flag': flag, 'help': help}.items():
        if v:
            metadata = metadata | {k: v}
    return field(  # type: ignore[call-overload]
        default=default,
        default_factory=factory,
        init=init,
        repr=repr,
        hash=hash,
        compare=compare,
        metadata=metadata,
    )


class _Opts(TypedDict, total=False):
    type: Required[Callable]
    nargs: str
    choices: Iterable


def _unwrap_type(tp: type) -> tuple[type, _Opts]:
    origin = get_origin(tp)
    args = get_args(tp)
    if not origin or not args:  # Not a generic type
        return tp, {'type': tp}

    if origin is list:  # `List[T]`
        cls, opts = _unwrap_type(args[0])
        return cls, {**opts, 'nargs': argparse.ZERO_OR_MORE}

    if (  # `Optional[T]` or `T | None`
        origin in {Union, types.UnionType}
        and len(args) == 2
        and types.NoneType in args
    ):
        [tp_] = set(args) - {types.NoneType}
        cls, opts = _unwrap_type(tp_)
        if opts.get('nargs') == argparse.ZERO_OR_MORE:
            return cls, opts
        return cls, {**opts, 'nargs': argparse.OPTIONAL}

    if origin is Literal:  # `Literal[x, y]`
        choices = get_args(tp)
        if len(tps := {type(c) for c in choices}) != 1:
            msg = f'Literal parameters should have the same type. Got: {tps}'
            raise TypeError(msg)
        [cls] = tps
        return cls, {'type': cls, 'choices': choices}

    msg = (
        'Only list, Optional and Literal are supported as generic types. '
        f'Got: {tp}'
    )
    raise TypeError(msg)


def _get_fields(fn: Callable) -> Iterator[Field]:
    if is_dataclass(fn):  # Shortcut
        yield from fields(fn)
        return

    for p in signature(fn).parameters.values():
        if p.kind is p.KEYWORD_ONLY and p.default is p.empty:
            msg = f'Keyword "{p.name}" must have default'
            raise ValueError(msg)
        if p.kind in {p.POSITIONAL_ONLY, p.VAR_POSITIONAL, p.VAR_KEYWORD}:
            msg = f'Unsupported parameter type: {p.kind}'
            raise TypeError(msg)

        if isinstance(p.default, Field):
            fd = p.default
        else:
            fd = arg(MISSING if p.default is p.empty else p.default)
        fd.name = p.name
        yield fd


def _get_metadata(tp: type, fd: Field) -> tuple[type, _Meta]:
    info = inspect_annotation(tp, annotation_source=AnnotationSource.CLASS)

    flag = fd.metadata.get('flag')
    name = fd.name.replace('_', '-')
    help_ = fd.metadata.get('help') or ''

    if info.type is not UNKNOWN:
        tp = info.type
        for m in info.metadata:
            if isinstance(m, Meta):
                help_ = m.help
                flag = m.flag

    return tp, _Meta(help=help_, flag=flag, name=name)


def _visit_nested(
    parser: ArgumentParser | _ArgumentGroup,
    fn: Callable,
    seen: dict[str, list],
) -> list[_Node]:
    try:
        hints = get_type_hints(fn, include_extras=True)
    except NameError:
        if fn.__module__ != '__main__':
            raise
        for finfo in stack():
            if not getmodule(f := finfo.frame):
                hints = get_type_hints(fn, f.f_globals, include_extras=True)
                break
        else:
            raise

    nodes: list[_Node] = []
    for fd in _get_fields(fn):
        if fd.init:
            seen.setdefault(fd.name, []).append(fn)
            nodes.append(_visit_field(parser, hints[fd.name], fd, seen))

    for name, usages in seen.items():
        if len(usages) > 1:
            msg = (
                f'Field name "{name}" occured multiple times: '
                + ', '.join(f'{c.__module__}.{c.__qualname__}' for c in usages)
                + '. All field names should be unique'
            )
            raise ValueError(msg)
    return nodes


def _visit_field(
    parser: ArgumentParser | _ArgumentGroup,
    tp: type,
    fd: Field,
    seen: dict[str, list],
) -> _Node:
    tp, meta = _get_metadata(tp, fd)
    cls, opts = _unwrap_type(tp)

    if cls is not bool and fd.default is not MISSING:
        meta.help += f' (default: {fd.default})'

    if is_dataclass(cls):  # Nested dataclass
        arg_group = parser.add_argument_group(fd.name)
        return fd.name, cls, _visit_nested(arg_group, cls, seen)

    vtp = opts['type']
    if (
        isinstance(vtp, type)
        and issubclass(vtp, Iterable)
        and not issubclass(vtp, str)
    ):
        msg = (
            'Iterable value types are supported only as generics. '
            f'Got: {vtp}'
        )
        raise TypeError(msg)

    flags = [meta.flag] if meta.flag else []
    default = (
        fd.default if fd.default_factory is MISSING else fd.default_factory()
    )

    if cls is bool:  # Optional
        if default is MISSING:
            msg = f'Boolean field "{fd.name}" should have default'
            raise ValueError(msg)
        parser.add_argument(
            f'--{meta.name}',
            *flags,
            action=BooleanOptionalAction,
            default=default,
            help=meta.help,
        )

    # Generic optional
    elif default is not MISSING:
        if opts.get('nargs') == argparse.OPTIONAL:
            del opts['nargs']
        parser.add_argument(
            f'--{meta.name}', *flags, **opts, default=default, help=meta.help
        )

    elif isinstance(parser, ArgumentParser):  # Allow only for root parser
        if meta.flag:
            msg = f'Positional-only field "{fd.name}" should not have flag'
            raise ValueError(msg)
        parser.add_argument(meta.name, **opts, help=meta.help)

    else:
        msg = (
            'Positional-only fields are forbidden for nested types. '
            f'Please set default value for "{fd.name}"'
        )
        raise ValueError(msg)

    return fd.name


def _construct[T](
    src: dict[str, Any], fn: Callable[..., T], args: Iterable[_Node]
) -> T:
    kwargs = {}
    for a in args:
        if isinstance(a, str):
            kwargs[a] = src.pop(a)
        else:
            kwargs[a[0]] = _construct(src, a[1], a[2])
    return fn(**kwargs)


def parse_args[T](
    fn: Callable[..., T],
    /,
    args: Sequence[str] | None = None,
    prog: str | None = None,
) -> tuple[T, ArgumentParser]:
    """Create parser from type hints of callable, parse args and do call."""
    # TODO: Rename to `run`
    if not callable(fn):
        raise TypeError(f'Expectet callable. Got: {type(fn).__qualname__}')

    parser = ArgumentParser(prog)
    nodes = _visit_nested(parser, fn, {})

    assert args is None or (
        isinstance(args, Sequence) and not isinstance(args, str)
    )

    namespace = parser.parse_args(args)
    obj = _construct(vars(namespace), fn, nodes)
    return obj, parser


def _import_from_string(qualname: str):
    modname, _, attrname = qualname.partition(':')
    if not modname or not attrname:
        msg = (
            f'Import string "{qualname}" must be '
            'in format "<module>:<attribute>".'
        )
        raise ImportError(msg)

    mod = importlib.import_module(modname)

    obj: Any = mod
    try:
        for a in attrname.split('.'):
            obj = getattr(obj, a)
    except AttributeError:
        msg = f'Attribute "{attrname}" not found in module "{modname}".'
        raise AttributeError(msg) from None
    return obj


if __name__ == '__main__':
    qualname, *argv = sys.argv
    obj = _import_from_string(qualname)
    parse_args(obj, argv)



__all__ = ['lock_seed', 'trace', 'trace_module', 'whereami']

import gc
import os
import random
from collections.abc import Callable, Iterator
from contextlib import suppress
from inspect import currentframe, getmodule, isfunction
from itertools import islice
from types import FrameType, ModuleType

import numpy as np
import wrapt

from ._cache import memoize
from ._import_hook import register_post_import_hook


def _frame_hash(frame: FrameType) -> tuple[str, int]:
    return frame.f_code.co_filename, frame.f_lineno


@memoize(100, policy='lru', key_fn=_frame_hash)
def _get_source(frame: FrameType) -> str:
    # Get source module name
    modname = (
        spec.name
        if (module := getmodule(frame)) and (spec := module.__spec__)
        else '__main__'
    )

    # Get source code name (method or function name)
    code = frame.f_code
    codename = next(
        (f.__qualname__ for f in gc.get_referrers(code) if isfunction(f)),
        code.co_name,
    )
    if codename == '<module>':
        codename = ''

    return f'{modname}:{codename}:{frame.f_lineno}'


def _get_source_calls(frame: FrameType | None) -> Iterator[str]:
    while frame:
        yield _get_source(frame)
        if frame.f_code.co_name == '<module>':  # Stop on module-level scope
            return
        frame = frame.f_back


def stack(skip: int = 0, limit: int | None = None) -> Iterator[str]:
    """Return iterator of FrameInfos, stopping on module-level scope."""
    frame = currentframe()
    calls = _get_source_calls(frame)
    calls = islice(calls, skip + 1, None)  # Skip 'skip' outerless frames
    if not limit:
        return calls
    return islice(calls, limit)  # Keep at most `limit` outer frames


def whereami(skip: int = 0, limit: int | None = None) -> str:
    calls = stack(skip + 1, limit)
    return ' -> '.join(reversed([*calls]))


@wrapt.decorator
def trace(fn, _, args, kwargs):
    print(
        f'<({whereami(3)})> : {fn.__module__ or ""}.{fn.__qualname__}',
        flush=True,
    )
    return fn(*args, **kwargs)


def _set_trace(
    obj: ModuleType | Callable,
    *,
    seen: set[str] | None = None,
    prefix: str | None = None,
    module: ModuleType | None = None,
) -> None:
    # TODO: rewrite using unittest.mock
    if isinstance(obj, ModuleType):
        if seen is None:
            seen = set()
            prefix = obj.__name__
        assert isinstance(prefix, str)
        if not obj.__name__.startswith(prefix) or obj.__name__ in seen:
            return
        seen.add(obj.__name__)
        for name in dir(obj):
            _set_trace(
                getattr(obj, name), module=obj, seen=seen, prefix=prefix
            )

    if not callable(obj):
        return

    assert isinstance(module, ModuleType)
    if not hasattr(obj, '__dict__'):
        setattr(module, obj.__qualname__, trace(obj))
        print(f'wraps "{module.__name__}:{obj.__qualname__}"')
        return

    for name in obj.__dict__:
        with suppress(AttributeError, TypeError):
            member = getattr(obj, name)
            if not callable(member):
                continue
            decorated = trace(member)

            for m in (decorated, member, obj):
                with suppress(AttributeError):
                    decorated.__module__ = m.__module__
                    break
            else:  # noqa: PLW0120, RUF100
                decorated.__module__ = getattr(module, '__name__', '')
            setattr(obj, name, decorated)
            print(f'wraps "{module.__name__}:{obj.__qualname__}.{name}"')


def trace_module(name: str) -> None:
    """Enable call logging for each callable inside module name."""
    register_post_import_hook(_set_trace, name)


# ---------------------------------------------------------------------------


def lock_seed(seed: int) -> None:
    """Set seed for all modules: random/numpy/torch."""
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)

    def _torch_seed(torch) -> None:
        torch.manual_seed(seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False

    register_post_import_hook(_torch_seed, 'torch')



from typing import Literal, overload

from ._futures import AnyBatchDecorator, PsAnyBatchDecorator
from ._types import CachePolicy, Decorator, KeyFn, PsDecorator

def cache_status() -> str: ...

# Unbound
@overload
def memoize(
    count: None = ...,
    *,
    policy: None = ...,
) -> Decorator: ...
@overload
def memoize[**P](
    count: None = ...,
    *,
    policy: None = ...,
    key_fn: KeyFn[P],
) -> PsDecorator[P]: ...

# Unbound batched
@overload
def memoize(
    count: None = ...,
    *,
    policy: None = ...,
    batched: Literal[True],
) -> AnyBatchDecorator: ...
@overload
def memoize[T](
    count: None = ...,
    *,
    policy: None = ...,
    batched: Literal[True],
    key_fn: KeyFn[T],
) -> PsAnyBatchDecorator[T]: ...

# Count-capped
@overload
def memoize(
    count: int,
    *,
    policy: CachePolicy = ...,
    key_fn: KeyFn = ...,
) -> Decorator: ...
@overload
def memoize[**P](
    count: int,
    *,
    policy: CachePolicy = ...,
    key_fn: KeyFn[P],
) -> PsDecorator[P]: ...

# Count-capped batched
@overload
def memoize(
    count: int,
    *,
    batched: Literal[True],
    policy: CachePolicy = ...,
) -> AnyBatchDecorator: ...
@overload
def memoize[T](
    count: int,
    *,
    batched: Literal[True],
    policy: CachePolicy = ...,
    key_fn: KeyFn[T],
) -> PsAnyBatchDecorator[T]: ...

# Byte-capped
@overload
def memoize(
    *,
    nbytes: int,
    policy: CachePolicy = ...,
) -> Decorator: ...
@overload
def memoize[**P](
    *,
    nbytes: int,
    policy: CachePolicy = ...,
    key_fn: KeyFn[P],
) -> PsDecorator[P]: ...

# Byte-capped batched
@overload
def memoize(
    *,
    nbytes: int,
    batched: Literal[True],
    policy: CachePolicy = ...,
) -> AnyBatchDecorator: ...
@overload
def memoize[T](
    *,
    nbytes: int,
    batched: Literal[True],
    policy: CachePolicy = ...,
    key_fn: KeyFn[T],
) -> PsAnyBatchDecorator[T]: ...



"""
? At each moment each thread serve one executor.
+ Threads are reused over all executors.
+ Idle threads are stopped on _TIMEOUT.
? Most recently used threads used first (a.k.a. stack, more dropout).
"""

__all__ = ['ThreadQuota']

import os
import sys
from collections import deque
from collections.abc import Callable
from concurrent.futures import Executor, Future
from concurrent.futures._base import LOGGER
from concurrent.futures.thread import _WorkItem
from queue import Empty, SimpleQueue
from threading import _register_atexit  # type: ignore[attr-defined]
from threading import Lock, Thread
from weakref import WeakSet

if sys.version_info >= (3, 14):
    from concurrent.futures.thread import WorkerContext

    _worker_ctx = WorkerContext(lambda: None, ())
else:
    _worker_ctx = None

# TODO: investigate hangups when _TIMEOUT <= .01
_TIMEOUT = 1
_MIN_IDLE = os.cpu_count() or 1

# ------------------------------- generics -----------------------------------


def _safe_call[**P, T](
    fn: Callable[P, T], *args: P.args, **kwargs: P.kwargs
) -> T | None:
    try:
        return fn(*args, **kwargs)
    except (Empty, IndexError, ValueError):
        return None


# ------------------------------ implementation ------------------------------

type _Pipe = SimpleQueue['ThreadQuota | None']

_shutdown = False  # set only by `_python_exit`
_shutdown_lock = Lock()  # Blocks worker creation on interpreter shutdown
_executors = WeakSet['ThreadQuota']()
_workers = WeakSet[Thread]()
_idle = deque[_Pipe]()


def _python_exit() -> None:
    global _shutdown  # noqa: PLW0603
    with _shutdown_lock:
        _shutdown = True

    for e in _executors:
        e.shutdown(cancel_futures=True)

    while q := _safe_call(_idle.pop):
        q.put(None)


_register_atexit(_python_exit)


def _worker(q: _Pipe) -> None:
    try:
        while executor := _safe_call(q.get, timeout=_TIMEOUT):
            while work_item := _safe_call(executor._work_queue.popleft):
                if sys.version_info >= (3, 14):
                    work_item.run(_worker_ctx)  # Process task
                else:
                    work_item.run()
                if _shutdown:
                    executor._shutdown = True
                    return

            executor._idle.append(1)  # Decrease worker usage for executor
            _idle.append(q)  # Mark worker as idle, LIFO/stack
            if _shutdown:
                return

    except BaseException:  # noqa: BLE001
        LOGGER.critical('Exception in worker', exc_info=True)
    finally:
        if _TIMEOUT:
            _safe_call(_idle.remove, q)  # Omit when '_idle' tracks weakrefs


class ThreadQuota(Executor):
    __slots__ = ('_idle', '_shutdown', '_shutdown_lock', '_work_queue')

    def __init__(self, max_workers: int) -> None:
        self._work_queue = deque[_WorkItem]()
        self._idle = [1] * max_workers  # semaphore

        self._shutdown_lock = Lock()
        self._shutdown = False

        with _shutdown_lock:
            _executors.add(self)

    def submit[**P, R](
        self, fn: Callable[P, R], /, *args: P.args, **kwargs: P.kwargs
    ) -> Future[R]:
        f = Future()  # type: ignore[var-annotated]
        self.submit_f(f, fn, *args, **kwargs)
        return f

    def submit_f[**P, R](
        self,
        f: Future[R],
        fn: Callable[P, R],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        with self._shutdown_lock, _shutdown_lock:
            if self._shutdown or _shutdown:
                msg = 'cannot schedule futures after shutdown'
                raise RuntimeError(msg)

            if sys.version_info >= (3, 14):
                self._work_queue.append(_WorkItem(f, (fn, args, kwargs)))
            else:
                self._work_queue.append(_WorkItem(f, fn, args, kwargs))

            if _safe_call(self._idle.pop):  # Pool is not maximized yet
                if q := _safe_call(_idle.pop):  # Use idle worker
                    q.put(self)
                else:  # Scale to new worker
                    q = SimpleQueue[ThreadQuota | None]()
                    q.put(self)
                    w = Thread(target=_worker, args=[q])
                    w.start()
                    _workers.add(w)

    def shutdown(
        self, wait: bool = True, *, cancel_futures: bool = False
    ) -> None:
        with self._shutdown_lock:
            if self._shutdown:
                return
            self._shutdown = True

            if cancel_futures:
                while work_item := _safe_call(self._work_queue.pop):
                    work_item.future.cancel()

            # TODO: if not `wait` - stop sub-workers
            if not _TIMEOUT:
                # Keep at most 25% of workers idle
                while len(_idle) > max(len(_workers) / 4, _MIN_IDLE) and (
                    q := _safe_call(_idle.popleft)
                ):
                    q.put(None)



__all__ = ['make_key']

from collections.abc import Hashable
from dataclasses import dataclass

_KWD_MARK = object()


@dataclass(frozen=True, slots=True)
class _HashedSeq:
    """Memorizes hash to not recompute it on cache search/update."""

    items: tuple
    hashvalue: int

    def __eq__(self, value: object) -> bool:
        return isinstance(value, _HashedSeq) and self.items == value.items

    def __hash__(self) -> int:
        return self.hashvalue


def make_key(*args, **kwargs) -> Hashable:
    """Make hashable key from function arguments.

    Copied from functools._make_key, as private function.
    """
    if kwargs:
        args = sum(kwargs.items(), (*args, _KWD_MARK))
    if len(args) == 1 and type(args[0]) in {int, str}:
        return args[0]
    return _HashedSeq(args, hash(args))



__all__ = ['as_actor', 'coroutine']

from collections import Counter, deque
from collections.abc import Callable, Generator, Hashable, Iterable, Iterator
from functools import update_wrapper
from threading import Lock

import wrapt

from ._more import _deiter


def coroutine[**P, Y, S, R](
    fn: Callable[P, Generator[Y, S, R]], /
) -> Callable[P, Generator[Y, S, R]]:
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Generator[Y, S, R]:
        coro = fn(*args, **kwargs)
        next(coro)
        return coro

    return update_wrapper(wrapper, fn)


class _Sync[Y, S, R](wrapt.ObjectProxy):  # type: ignore[misc]
    __wrapped__: Generator[Y, S, R]

    def __init__(self, wrapped: Generator[Y, S, R]) -> None:
        super().__init__(wrapped)
        self._self_lock = Lock()

    def _call[**P, T](
        self, op: Callable[P, T], /, *args: P.args, **kwargs: P.kwargs
    ) -> T:
        with self._self_lock:
            return op(*args, **kwargs)

    def __next__(self) -> Y:
        return self._call(self.__wrapped__.__next__)

    def send(self, item: S, /) -> Y:
        return self._call(self.__wrapped__.send, item)

    def throw(self, value: BaseException, /) -> Y:
        return self._call(self.__wrapped__.throw, value)

    def close(self) -> None:
        self._call(self.__wrapped__.close)


def threadsafe_iter[**P, Y, S, R](
    fn: Callable[P, Generator[Y, S, R]], /
) -> Callable[P, Generator[Y, S, R]]:
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Generator[Y, S, R]:
        gen = fn(*args, **kwargs)
        return _Sync(gen)

    return update_wrapper(wrapper, fn)


@threadsafe_iter
@coroutine
def summary() -> Generator[None, Hashable | None]:
    # ? delete this or find use case
    state = Counter[Hashable]()
    while True:
        key = yield
        if key is None:
            state.clear()
        else:
            state[key] += 1
            print(dict(state), flush=True, end='\r')


@threadsafe_iter
@coroutine
def as_actor[T, R](
    transform: Callable[[Iterable[T]], Iterator[R]],
) -> Generator[R, T]:
    buf = deque[T]()
    gen = transform(_deiter(buf))  # infinite

    # shortcuts
    buf_append, gen_next = buf.append, gen.__next__

    x = yield  # type: ignore[misc]  # preseed coroutine
    while True:
        buf_append(x)
        x = yield gen_next()



__all__ = ['aceil', 'afloor', 'apack', 'around', 'pascal', 'smallest_dtype']

import numpy as np
import numpy.typing as npt


def _force_int(
    op: np.ufunc, a: npt.NDArray[np.number], dtype: npt.DTypeLike
) -> npt.NDArray[np.integer]:
    dtype = np.dtype(dtype)
    if dtype.kind not in 'iu':
        msg = f'Cannot cast to non-integer dtype: {dtype}'
        raise ValueError(msg)

    match a.dtype.kind:
        case 'b' | 'u' | 'i':
            return a.astype(dtype)
        case 'f':
            return op(a, out=np.empty_like(a, dtype), casting='unsafe')
        case _:
            msg = f'Unsupported dtype: {a.dtype}'
            raise ValueError(msg)


def around(
    x: npt.NDArray[np.number], dtype: npt.DTypeLike = int
) -> npt.NDArray[np.integer]:
    """Faster alternative to `np.round(x).astype(int)`."""
    return _force_int(np.rint, x, dtype)


def aceil(
    x: npt.NDArray[np.number], dtype: npt.DTypeLike = int
) -> npt.NDArray[np.integer]:
    """Faster alternative to `np.ceil(x).astype(int)`."""
    return _force_int(np.ceil, x, dtype)


def afloor(
    x: npt.NDArray[np.number], dtype: npt.DTypeLike = int
) -> npt.NDArray[np.integer]:
    """Faster alternative to `np.floor(x).astype(int)`."""
    return _force_int(np.floor, x, dtype)


def smallest_dtype(
    a_min: float | int, a_max: float | int | None = None
) -> np.dtype:
    r = np.min_scalar_type(a_min)
    return r if a_max is None else np.result_type(r, a_max)


def apack(
    a: npt.ArrayLike | npt.NDArray[np.integer],
    a_min: int | None = None,
    a_max: int | None = None,
) -> npt.NDArray[np.integer]:
    """Convert integer array to smallest dtype."""
    a = np.asarray(a)
    if a.dtype.kind != 'i':
        msg = f'Cannot pack non-integer array: {a.dtype}'
        raise ValueError(msg)
    if not a.size:
        return a

    if a_min is None:
        a_min = a.min()
        assert a_min is not None

    if a_max is None:
        a_max = a.max()

    if (dtype := smallest_dtype(a_min, a_max)).itemsize < a.itemsize:
        return a.astype(dtype)
    return a


def pascal(n: int) -> npt.NDArray[np.int64]:
    if n < 1 or n > 67:
        msg = '`n` must in 1..67 range'
        raise ValueError(msg)
    a = np.zeros(n, 'q')
    a[0] = 1
    for _ in range(n - 1):
        a[1:] += a[:-1]
    return a



from collections.abc import Callable
from dataclasses import dataclass
from pathlib import Path
from typing import Annotated, Any, Literal

import pytest

from glow.cli import Meta, parse_args


@dataclass
class Positional:
    arg: str


@dataclass
class UntypedList:  # Forbidden, as list field should always be typed
    args: list


@dataclass
class List_:  # noqa: N801
    args: list[str]


@dataclass
class UnsupportedTuple:
    args: tuple[str, int]


@dataclass
class UnsupportedSet:
    args: set[str]


@dataclass
class BadBoolean:  # Forbidden, as boolean field should always have default
    arg: bool


@dataclass
class Boolean:
    param: bool = False


@dataclass
class Nullable:
    param: str | None = None


@dataclass
class Optional_:  # noqa: N801
    param: str = 'hello'


@dataclass
class Nested:
    arg: str
    nested: Optional_


@dataclass
class NestedPositional:  # Forbidden, as only top level args can be positional
    arg2: str
    nested: Positional


@dataclass
class Aliased:
    arg: str = 'hello'


@dataclass
class NestedAliased:  # Forbidden as all field names must be unique
    arg: str
    nested: Aliased


@dataclass
class Custom:
    arg: Path


@dataclass
class AnnotatedPositional:
    arg: Annotated[int, Meta(help='help')]


@dataclass
class FlagKeyword:
    param: Annotated[int, Meta(help='help', flag='-p')] = 42


@dataclass
class AnnotatedKeyword:
    param: Annotated[int, Meta(help='help')] = 42


@pytest.mark.parametrize(
    ('argv', 'expected'),
    [
        (['value'], Positional('value')),
        ([], List_([])),
        (['a'], List_(['a'])),
        (['a', 'b'], List_(['a', 'b'])),
        ([], Boolean()),
        (['--no-param'], Boolean()),
        (['--param'], Boolean(True)),
        ([], Nullable()),
        (['--param', 'value'], Nullable('value')),
        ([], Optional_()),
        (['--param', 'world'], Optional_('world')),
        (['value'], Nested('value', Optional_())),
        (['value', '--param', 'pvalue'], Nested('value', Optional_('pvalue'))),
        (['test.txt'], Custom(Path('test.txt'))),
        (['5'], AnnotatedPositional(5)),
        (['--param', '5'], AnnotatedKeyword(5)),
        ([], AnnotatedKeyword(42)),
        (['-p', '42'], FlagKeyword(42)),
    ],
)
def test_good_class(argv: list[str], expected: Any):
    cls = type(expected)
    result, _ = parse_args(cls, argv)
    assert isinstance(result, cls)
    assert result == expected


@pytest.mark.parametrize(
    ('cls', 'exc_type'),
    [
        (Positional, SystemExit),
        (BadBoolean, ValueError),
        (UnsupportedTuple, TypeError),
        (UnsupportedSet, TypeError),
        (UntypedList, TypeError),
        (Nested, SystemExit),
        (NestedPositional, ValueError),
        (NestedAliased, ValueError),
    ],
)
def test_bad_class(cls: type[Any], exc_type: type[BaseException]):
    with pytest.raises(exc_type):
        parse_args(cls, [])


def _no_op():
    return ()


def _positional(a: int):
    return a


def _keyword(a: int = 4):
    return a


def _kw_nullable(a: int = None):  # type: ignore[assignment]  # noqa: RUF013
    return a


def _kwarg_literal(a: Literal[1, 2] = 1):
    return a


def _kwarg_bool(a: bool = False):
    return a


def _kwarg_list(a: list[int] = []):  # noqa: B006
    return a


def _kwarg_opt_list(a: list[int] | None = None):
    return a


def _arg_kwarg(a: int, b: str = 'hello'):
    return a, b


@pytest.mark.parametrize(
    ('argv', 'func', 'expected'),
    [
        ([], _no_op, ()),
        (['42'], _positional, 42),
        ([], _keyword, 4),
        (['--a', '58'], _keyword, 58),
        ([], _kw_nullable, None),
        (['--a', '73'], _kw_nullable, 73),
        ([], _kwarg_literal, 1),
        (['--a', '2'], _kwarg_literal, 2),
        ([], _kwarg_bool, False),
        (['--no-a'], _kwarg_bool, False),
        (['--a'], _kwarg_bool, True),
        ([], _kwarg_list, []),
        (['--a'], _kwarg_list, []),
        (['--a', '1'], _kwarg_list, [1]),
        (['--a', '1', '2'], _kwarg_list, [1, 2]),
        ([], _kwarg_opt_list, None),
        (['--a'], _kwarg_opt_list, []),
        (['--a', '1'], _kwarg_opt_list, [1]),
        (['--a', '1', '2'], _kwarg_opt_list, [1, 2]),
        (['53'], _arg_kwarg, (53, 'hello')),
        (['87', '--b', 'bye'], _arg_kwarg, (87, 'bye')),
    ],
)
def test_good_func[T](argv: list[str], func: Callable[..., T], expected: T):
    result, _ = parse_args(func, argv)
    assert type(result) is type(expected)
    assert result == expected



__all__ = ['imhash_hist']

from pathlib import Path

import cv2
import numpy as np
import numpy.typing as npt
from PIL.Image import Image

type _U8 = npt.NDArray[np.uint8]
type _F32 = npt.NDArray[np.float32]
type _AnyImage = Path | str | Image | _F32 | _U8


def imhash_hist(x: _AnyImage, /, *, bins: int = 4) -> _F32 | None:
    """Compute histogram-based L2-normalized image hash.

    To measure distance use dot product (i.e. `hash1 @ hash2`).
    Large dot product means images have similar color distribution.
    """
    image = _ensure_image(x)
    if image is None:
        return None

    # By default 4x4x4, i.e. 64 values
    hist = cv2.calcHist(
        [image],
        channels=[0, 1, 2],
        mask=None,
        histSize=[bins, bins, bins],
        ranges=[0, 256, 0, 256, 0, 256],
    )
    hist = np.float32(hist).ravel()
    l2 = cv2.norm(hist, normType=cv2.NORM_L2)
    return hist / l2


def _ensure_image(i: _AnyImage, /) -> _U8 | None:
    match i:
        case Path() | str():
            return cv2.imread(str(i), cv2.IMREAD_COLOR)  # type: ignore
        case Image():
            return np.asarray(i.convert('RGB'), 'B')
        case np.ndarray():
            i = _ensure_u8(i)
            return _ensure_rgb(i)
        case _:
            return None


def _ensure_rgb(image: _U8, /) -> _U8:
    match image.shape:
        case (_, _) | (_, _, 1):
            return cv2.cvtColor(image, cv2.COLOR_GRAY2RGB)  # type: ignore
        case (_, _, 3):
            return image
        case _ as unsupported:
            msg = f'Input is not image. Got shape {unsupported}'
            raise ValueError(msg)


def _ensure_u8(image: _F32 | _U8, /) -> _U8:
    dt = image.dtype
    if dt == 'B':
        return image  # type: ignore
    if dt == 'f':
        return (image * 255).clip(0, 255).astype('B')  # assume [0 .. 1]
    msg = f'Unsupported image dtype. Got {dt}'
    raise NotImplementedError(msg)



from collections.abc import Callable, Iterable, Iterator, Mapping
from concurrent.futures import Executor
from contextlib import AbstractContextManager
from typing import TypedDict, Unpack, overload

class _MapKwargs(TypedDict, total=False):
    max_workers: int | None
    prefetch: int | None
    mp: bool
    chunksize: int | None

class _MapIterKwargs(_MapKwargs, total=False):
    unordered: bool

def max_cpu_count(upper_bound: int = ..., *, mp: bool = ...) -> int: ...
def get_executor(
    max_workers: int, mp: bool
) -> AbstractContextManager[Executor]: ...
def buffered[T](
    __iter: Iterable[T], /, *, latency: int = ..., mp: bool | Executor = ...
) -> Iterator[T]: ...
def starmap_n[R](
    __func: Callable[..., R],
    __iter: Iterable[Iterable],
    /,
    **kwargs: Unpack[_MapIterKwargs],
) -> Iterator[R]: ...
@overload
def map_n[T, R](
    __func: Callable[[T], R],
    __iter1: Iterable[T],
    /,
    **kwargs: Unpack[_MapIterKwargs],
) -> Iterator[R]: ...
@overload
def map_n[T1, T2, R](
    __f: Callable[[T1, T2], R],
    __iter1: Iterable[T1],
    __iter2: Iterable[T2],
    /,
    **kwargs: Unpack[_MapIterKwargs],
) -> Iterator[R]: ...
@overload
def map_n[T1, T2, T3, R](
    __f: Callable[[T1, T2, T3], R],
    __iter1: Iterable[T1],
    __iter2: Iterable[T2],
    __iter3: Iterable[T3],
    /,
    **kwargs: Unpack[_MapIterKwargs],
) -> Iterator[R]: ...
@overload
def map_n[R](
    __func: Callable[..., R],
    __iter1: Iterable,
    __iter2: Iterable,
    __iter3: Iterable,
    __iter4: Iterable,
    /,
    *__iters: Iterable,
    **kwargs: Unpack[_MapIterKwargs],
) -> Iterator[R]: ...
def map_n_dict[T, K, R](
    func: Callable[[T], R], obj: Mapping[K, T], /, **kwargs: Unpack[_MapKwargs]
) -> dict[K, R]: ...



from collections.abc import AsyncIterator, Callable, Mapping
from contextlib import AbstractAsyncContextManager
from typing import Any, Required, TypedDict, Unpack, overload

from ._futures import ABatchDecorator, ABatchFn
from ._types import AnyIterable, Coro

class _AmapKwargs(TypedDict, total=False):
    limit: Required[int]
    unordered: bool

def astarmap[*Ts, R](
    func: Callable[[*Ts], Coro[R]],
    iterable: AnyIterable[tuple[*Ts]],
    /,
    **kwargs: Unpack[_AmapKwargs],
) -> AsyncIterator[R]: ...
@overload
def amap[T, R](
    func: Callable[[T], Coro[R]],
    iter1: AnyIterable[T],
    /,
    **kwargs: Unpack[_AmapKwargs],
) -> AsyncIterator[R]: ...
@overload
def amap[T, T2, R](
    func: Callable[[T, T2], Coro[R]],
    iter1: AnyIterable[T],
    iter2: AnyIterable[T2],
    /,
    **kwargs: Unpack[_AmapKwargs],
) -> AsyncIterator[R]: ...
@overload
def amap[T, T2, T3, R](
    func: Callable[[T, T2, T3], Coro[R]],
    iter1: AnyIterable[T],
    iter2: AnyIterable[T2],
    iter3: AnyIterable[T3],
    /,
    **kwargs: Unpack[_AmapKwargs],
) -> AsyncIterator[R]: ...
@overload
def amap[T, T2, T3, T4, R](
    func: Callable[[T, T2, T3, T4], Coro[R]],
    iter1: AnyIterable[T],
    iter2: AnyIterable[T2],
    iter3: AnyIterable[T3],
    iter4: AnyIterable[T4],
    /,
    **kwargs: Unpack[_AmapKwargs],
) -> AsyncIterator[R]: ...
@overload
def amap[R](
    func: Callable[..., Coro[R]],
    iter1: AnyIterable,
    iter2: AnyIterable,
    iter3: AnyIterable,
    iter4: AnyIterable,
    iter5: AnyIterable,
    /,
    *iters: AnyIterable,
    **kwargs: Unpack[_AmapKwargs],
) -> AsyncIterator[R]: ...
async def amap_dict[K, T1, T2](
    func: Callable[[T1], Coro[T2]], obj: Mapping[K, T1], /, *, limit: int
) -> dict[K, T2]: ...
@overload
def azip() -> AsyncIterator[Any]: ...
@overload
def azip[T](
    iter1: AnyIterable[T], /
) -> AsyncIterator[tuple[T]]: ...  # noqa: RUF100,Y090
@overload
def azip[T, T2](
    iter1: AnyIterable[T], iter2: AnyIterable[T2], /
) -> AsyncIterator[tuple[T, T2]]: ...
@overload
def azip[T, T2, T3](
    iter1: AnyIterable[T], iter2: AnyIterable[T2], iter3: AnyIterable[T3], /
) -> AsyncIterator[tuple[T, T2, T3]]: ...
@overload
def azip[T, T2, T3, T4](
    iter1: AnyIterable[T],
    iter2: AnyIterable[T2],
    iter3: AnyIterable[T3],
    iter4: AnyIterable[T4],
    /,
) -> AsyncIterator[tuple[T, T2, T3, T4]]: ...
@overload
def azip(
    iter1: AnyIterable,
    iter2: AnyIterable,
    iter3: AnyIterable,
    iter4: AnyIterable,
    iter5: AnyIterable,
    /,
    *iters: AnyIterable,
) -> AsyncIterator[tuple]: ...
@overload
def astreaming(
    *, batch_size: int | None = ..., timeout: float = ...
) -> ABatchDecorator: ...
@overload
def astreaming[T, R](
    fn: ABatchFn[T, R],
    /,
    *,
    batch_size: int | None = ...,
    timeout: float = ...,
) -> ABatchFn[T, R]: ...

class RwLock:
    def __init__(self) -> None: ...
    def read(self) -> AbstractAsyncContextManager: ...
    def write(self) -> AbstractAsyncContextManager: ...



import asyncio
import random
import time
from threading import Event, Thread
from unittest.mock import MagicMock

import pytest

import glow


def _lrange(*args: int) -> list[int]:
    return list(range(*args))


def test_bad_len():
    @glow.memoize(5, batched=True)
    def fn(xs):
        return []

    with pytest.raises(RuntimeError):
        fn([None])
    assert not dict(fn.cache)


@pytest.mark.asyncio
async def test_bad_len_async():
    @glow.memoize(5, batched=True)
    async def fn(xs):
        return []

    with pytest.raises(RuntimeError):
        await fn([None])
    assert not dict(fn.cache)


def test_seq():
    load = MagicMock(side_effect=lambda x: x)

    @glow.memoize(20, batched=True)
    def load_batch(xs):
        return [load(x) for x in xs]

    for _ in range(5):
        numbers = random.choices(range(20), k=5)
        results = load_batch(numbers)
        assert numbers == results

    called_with = (x for (x,), _ in load.call_args_list)
    assert sorted(called_with) == sorted(load_batch.cache)


@pytest.mark.asyncio
async def test_seq_async():
    load = MagicMock(side_effect=lambda x: x)

    @glow.memoize(20, batched=True)
    async def load_batch(xs):
        await asyncio.sleep(0)
        return [load(x) for x in xs]

    for _ in range(5):
        numbers = random.choices(range(20), k=5)
        results = await load_batch(numbers)
        assert numbers == results

    called_with = (x for (x,), _ in load.call_args_list)
    assert sorted(called_with) == sorted(load_batch.cache)


@pytest.mark.parametrize('max_workers', [2, 4] * 10)
def test_map(max_workers):
    @glow.memoize(100, batched=True)
    def load_batch(xs):
        return [*xs]

    numbers = random.choices(range(100), k=100)
    nchunks = glow.chunked(numbers, 5)
    rchunks = glow.map_n(load_batch, nchunks, max_workers=max_workers)
    assert [r for c in rchunks for r in c] == numbers


@pytest.mark.asyncio
@pytest.mark.parametrize('max_workers', [2, 4] * 10)
async def test_amap(max_workers):
    @glow.memoize(100, batched=True)
    async def load_batch(xs):
        await asyncio.sleep(0)
        return [*xs]

    numbers = random.choices(range(100), k=100)
    nchunks = glow.chunked(numbers, 5)
    rchunks = [
        x async for x in glow.amap(load_batch, nchunks, limit=max_workers)
    ]
    assert [r for c in rchunks for r in c] == numbers


def test_concurrent():
    ev = Event()

    @glow.memoize(100, batched=True)
    def fn(xs):
        ev.set()
        time.sleep(0.1)
        return xs

    t = Thread(target=lambda: fn(_lrange(0, 10)))
    t.start()
    ev.wait()

    x1 = fn(_lrange(5, 15))
    t.join()

    assert x1 == _lrange(5, 15)
    assert dict(fn.cache).keys() == set(_lrange(15))


@pytest.mark.asyncio
async def test_concurrent_async():
    ev = asyncio.Event()

    @glow.memoize(100, batched=True)
    async def fn(xs):
        ev.set()
        await asyncio.sleep(0.1)
        return xs

    t = asyncio.create_task(fn(_lrange(0, 10)))
    await ev.wait()

    x1 = await fn(_lrange(5, 15))
    assert x1 == _lrange(5, 15)

    x2 = await t
    assert x2 == _lrange(0, 10)

    assert dict(fn.cache).keys() == set(_lrange(15))


@pytest.mark.asyncio
async def test_concurrent_async_interrupted():
    ev = asyncio.Event()

    @glow.memoize(100, batched=True)
    async def fn(xs):
        ev.set()
        await asyncio.sleep(0.2)
        return xs

    t = asyncio.create_task(fn(_lrange(10)))
    await ev.wait()

    with pytest.raises(asyncio.TimeoutError):
        async with asyncio.timeout(0.1):
            await fn(_lrange(5, 15))

    await t
    assert dict(fn.cache).keys() == set(_lrange(10))



__all__ = [
    'as_iter',
    'chunked',
    'eat',
    'groupby',
    'ichunked',
    'ilen',
    'roundrobin',
    'windowed',
]

import threading
from collections import deque
from collections.abc import (
    Callable,
    Hashable,
    Iterable,
    Iterator,
    Mapping,
    Sequence,
    Sized,
)
from itertools import batched, chain, count, cycle, islice, repeat
from typing import Protocol, overload


class SupportsSlice[T](Sized, Protocol):
    def __getitem__(self, s: slice, /) -> T: ...


# ----------------------------------------------------------------------------


def as_iter[T](
    obj: Iterable[T] | T, /, limit: int | None = None
) -> Iterator[T]:
    """Make iterator with at most `limit` items."""
    if isinstance(obj, Iterable):
        return islice(obj, limit)
    return repeat(obj) if limit is None else repeat(obj, limit)


# ----------------------------------------------------------------------------


@overload
def _dispatch[S, *Ts](
    fallback_fn,
    fn: Callable[[SupportsSlice[S], *Ts], Iterator[S]],
    it: SupportsSlice[S],
    *args: *Ts,
) -> Iterator[S]: ...


@overload
def _dispatch[T, *Ts](
    fallback_fn: Callable[[Iterable[T], *Ts], Iterator[T]],
    fn,
    it: Iterable[T],
    *args: *Ts,
) -> Iterator[tuple[T, ...]]: ...


def _dispatch(fallback_fn, fn, it, *args):
    if (
        not isinstance(it, Sized)
        or not hasattr(it, '__getitem__')
        or isinstance(it, Mapping)
    ):
        return fallback_fn(it, *args)

    r = fn(it, *args)
    if isinstance(it, Sequence):
        return r

    try:
        # Ensure that slice is supported by prefetching 1st item
        first_or_none = tuple(islice(r, 1))
    except TypeError:
        return fallback_fn(it, *args)
    else:
        return chain(first_or_none, r)


# ----------------------------------------------------------------------------


def window_hint(it: Sized, size) -> int:
    return len(it) + 1 - size


def chunk_hint(it: Sized, size: int) -> int:
    return len(range(0, len(it), size))


def _sliced_windowed[T](s: SupportsSlice[T], size: int, /) -> Iterator[T]:
    indices = range(len(s) + 1)
    slices = map(slice, indices[:-size], indices[size:])
    return map(s.__getitem__, slices)


def _windowed[T](it: Iterable[T], size: int, /) -> Iterator[tuple[T, ...]]:
    if size == 1:  # Trivial case
        return zip(it)

    it = iter(it)
    w = deque(islice(it, size), maxlen=size)

    if len(w) != size:
        return iter(())
    return map(tuple, chain([w], map(w.__iadd__, zip(it))))


def _sliced[T](s: SupportsSlice[T], size: int, /) -> Iterator[T]:
    indices = range(len(s) + size)
    slices = map(slice, indices[::size], indices[size::size])
    return map(s.__getitem__, slices)


def _chunked[T](it: Iterable[T], size: int, /) -> Iterator[tuple[T, ...]]:
    if size == 1:  # Trivial case
        return zip(it)
    return batched(it, size, strict=False)


# ---------------------------------------------------------------------------


@overload
def windowed[T](it: SupportsSlice[T], size: int, /) -> Iterator[T]: ...


@overload
def windowed[T](it: Iterable[T], size: int, /) -> Iterator[tuple[T, ...]]: ...


def windowed(it, size, /):
    """Retrieve overlapped windows from iterable.

    Tries to use slicing if possible.

    >>> [*windowed(range(6), 3)]
    [range(0, 3), range(1, 4), range(2, 5), range(3, 6)]

    >>> [*windowed(iter(range(6)), 3)]
    [(0, 1, 2), (1, 2, 3), (2, 3, 4), (3, 4, 5)]
    """
    return _dispatch(_windowed, _sliced_windowed, it, size)


@overload
def chunked[T](__it: SupportsSlice[T], size: int, /) -> Iterator[T]: ...


@overload
def chunked[T](__it: Iterable[T], size: int, /) -> Iterator[tuple[T, ...]]: ...


def chunked(it, size):
    """Split iterable to chunks of at most size items each.

    Uses slicing if possible.
    Each next() on result will advance passed iterable to size items.

    >>> [*chunked(range(10), 3)]
    [range(0, 3), range(3, 6), range(6, 9), range(9, 10)]

    >>> [*chunked(iter(range(10)), 3)]
    [(0, 1, 2), (3, 4, 5), (6, 7, 8), (9,)]
    """
    return _dispatch(_chunked, _sliced, it, size)


# ----------------------------------------------------------------------------


def _deiter[T](q: deque[T], /) -> Iterator[T]:
    # Same as iter_except(q.popleft, IndexError) from docs of itertools
    try:
        while True:
            yield q.popleft()
    except IndexError:
        return


def ichunked[T](it: Iterable[T], size: int, /) -> Iterator[Iterator[T]]:
    """Split iterable to chunks of at most size items each.

    Does't consume items from passed iterable to return complete chunk
    unlike chunked, as yields iterators, not sequences.

    >>> s = ichunked(range(10), 3)
    >>> len(s)
    4
    >>> [[*chunk] for chunk in s]
    [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9]]
    """
    if size == 1:  # Trivial case
        yield from map(iter, zip(it))
        return

    it = iter(it)
    while head := deque(islice(it, 1)):
        # Remaining chunk items
        body = islice(it, size - 1)

        # Cache for not-yet-consumed
        tail = deque[T]()

        # Include early fetched item into chunk
        yield chain(_deiter(head), body, _deiter(tail))

        # Advance and fill internal cache, expand tail with items from body
        tail += body


# ----------------------------------------------------------------------------


def ilen(iterable: Iterable, /) -> int:
    """Return number of items in *iterable*.

    This consumes iterable, so handle with care.
    See `more_itertools.ilen`.
    """
    counter = count()
    deque(zip(iterable, counter), maxlen=0)
    return next(counter)


def eat(iterable: Iterable, /, *, daemon: bool = False) -> None:
    """Consume iterable, daemonize if needed (move to background thread)."""
    if daemon:
        threading.Thread(target=deque, args=(iterable, 0), daemon=True).start()
    else:
        deque(iterable, 0)  # Same as `more_itertools.consume(..., n=None)`


def roundrobin[T](*iterables: Iterable[T]) -> Iterator[T]:
    """roundrobin('ABC', 'D', 'EF') --> A D E B F C"""
    iters = cycle(iter(it) for it in iterables)
    for pending in range(len(iterables) - 1, -1, -1):
        yield from map(next, iters)
        iters = cycle(islice(iters, pending))


# ----------------------------------------------------------------------------


@overload
def groupby[T, K: Hashable](
    iterable: Iterable[T], /, key: Callable[[T], K]
) -> dict[K, list[T]]: ...


@overload
def groupby[T, K: Hashable, V](
    iterable: Iterable[T], /, key: Callable[[T], K], value: Callable[[T], V]
) -> dict[K, list[V]]: ...


def groupby(iterable, /, key, value=lambda x: x):
    """Group items from iterable by key.

    >>> groupby([True, (), 1, 0], bool)
    {True: [True, 1], False: [(), 0]}

    """
    r: dict = {}
    for x in iterable:
        r.setdefault(key(x), []).append(value(x))
    return r



__all__ = ['Stream', 'cumsum', 'maximum_cumsum']

from collections import deque
from dataclasses import dataclass
from itertools import accumulate

from ._types import Callback, Get


@dataclass(frozen=True, slots=True, repr=False)
class Stream[Y, S]:
    init: S
    push: Callback[S]
    pop: Get[Y]

    def send(self, value: S) -> Y:
        self.push(value)
        return self.pop()

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.send(self.init)})'


def cumsum() -> Stream[int, int]:
    """Stream running cumulative sum.

    Coroutine version of:
        >>> numbers = [-1, -2, 3, -4, 5, 7]
        ... np.cumsum(numbers)
        [-1, -3, 0, -4, 1, 8]

    Usage:
        >>> m = cumsum()
        ... numbers = [-1, -2, 3, -4, 5, 7]
        ... [m.send(x) for x in numbers]
        [-1, -3, 0, -4, 1, 8]
    """
    todo = deque[int]()

    values = iter(todo.popleft, None)
    partial_sums = accumulate(values)

    return Stream(init=0, push=todo.append, pop=partial_sums.__next__)


def maximum_cumsum() -> Stream[int, int]:
    """Stream running maximum cumulative sum.

    Coroutine version of:
        >>> numbers = [1, -1, 1, 1, -1, -1]
        ... np.maximum.accumulate(np.cumsum(numbers))
        [1, 1, 1, 2, 2, 2]

    Usage:
        >>> m = maximum_cumsum()
        ... numbers = [1, -1, 1, 1, -1, -1]
        ... [m.send(x) for x in numbers]
        [1, 1, 1, 2, 2, 2]
    """
    todo = deque[int]()

    values = iter(todo.popleft, None)
    partial_sums = accumulate(values)
    max_partial_sums = accumulate(partial_sums, max)

    return Stream(init=0, push=todo.append, pop=max_partial_sums.__next__)



"""IceCream - Never use print() to debug again.

Ansgar Grunseid
grunseid.com
grunseid@gmail.com

Pavel Maevskikh
arquolo@gmail.com

License: MIT

pip install asttokens colorama executing numpy pygments
"""

__all__ = ['ic', 'ic_repr']

import ast
import inspect
import pprint
import shutil
import sys
from collections.abc import Iterable, Iterator, Mapping
from dataclasses import fields, is_dataclass, replace
from datetime import datetime
from os.path import basename
from textwrap import dedent
from threading import Lock
from types import FrameType
from typing import Any, NamedTuple, overload

import executing
import numpy as np
from executing.executing import EnhancedAST
from pygments import highlight
from pygments.formatters import TerminalFormatter
from pygments.lexers.python import PythonLexer

if sys.platform == 'win32':
    import colorama

    colorama.init()

LOCK = Lock()

PREFIX = 'ic| '
LINE_WRAP_WIDTH = 70  # Characters

FORMATTER = TerminalFormatter(bg='dark')
LEXER = PythonLexer(ensurenl=False)


def is_literal(s) -> bool:
    try:
        ast.literal_eval(s)
    except Exception:  # noqa: BLE001
        return False
    else:
        return True


class Source(executing.Source):
    def get_text_with_indentation(self, node) -> str:
        result = self.asttokens().get_text(node)
        if '\n' in result:
            result = ' ' * node.first_token.start[1] + result
            result = dedent(result)
        return result.strip()


def indented_lines(prefix: str, lines: str) -> list[str]:
    space = ' ' * len(prefix)
    first, *rest = lines.splitlines()
    return [prefix + first] + [space + line for line in rest]


def format_pair(arg: str, value: str) -> str:
    # Align the start of multiline strings.
    if value[0] + value[-1] in ["''", '""']:
        value = ' '.join(value.splitlines(keepends=True))

    *lines, tail = arg.splitlines()
    return '\n'.join(lines + indented_lines(tail + ': ', value))


def _get_nd_grad(arr: np.ndarray) -> np.ndarray:
    # A bit sophisticated way to compute gradients by all directions,
    # but a fastest one.
    # Split tensor by all axes, do mean for each cell,
    # and then aggregate means to mean for the each axis split.
    arr = arr.squeeze()
    hs, ms = np.divmod(arr.shape, 2)

    # Pyramid of splits
    splits: dict[tuple[int, ...], np.ndarray] = {(): arr.astype('f')}
    for axis, (half, m) in enumerate(zip(hs, ms)):
        sep = [half, half + 1] if m else [half]
        splits = {
            (*k, k2): ss
            for k, s in splits.items()
            for k2, ss in enumerate(np.split(s, sep, axis))
        }

    # Tensor of means, (low, 0?, high) ^ ndim
    s_shape = 2 * hs.clip(max=1) + ms
    sums = np.zeros(s_shape)
    counts = np.zeros(s_shape, 'i')
    for loc, s in splits.items():
        n = s.size
        if (mask := np.ma.getmask(s)) is not np.ma.nomask:
            n -= mask.sum()
        if n:
            sums[loc] = s.sum()
        counts[loc] = n

    # Aggregate and do grads
    hsums = np.empty((arr.ndim, 2), 'f')
    for axis in range(arr.ndim):
        axes = *range(axis), *range(axis + 1, arr.ndim)
        hsums[axis] = (sums.sum(axes) / counts.sum(axes).clip(min=1))[[0, -1]]
    return hsums @ [-1, 1]


_MAX_SIZES = {'b': 40, 'i': 40, 'u': 40, 'f': 20, 'c': 20}


def _fmt_1d(a: np.ndarray, tol: int = 4) -> str:
    if a.size == 1:
        return f'{a.item():.{tol}g}'
    return '[' + ' '.join(f'{x:.{tol}g}' for x in a.tolist()) + ']'


def _bool_info(arr: np.ndarray) -> Iterator[str]:
    if arr.size < 500:  # all data in Packed Hex
        line = np.packbits(arr.flat).tobytes().hex().replace('0', '_')
        yield f'bits={line!r}'

    else:  # histrogram + gradient
        weights = np.bincount(arr.ravel()).astype('d') / arr.size
        yield f'bool @ {_fmt_1d(weights)}'
        yield from _grad_info(arr)


def _int_info(arr: np.ndarray, lo, hi, dtype) -> Iterator[str]:
    range_ = int(hi) - int(lo) + 1

    # Small range (lo >= 0 and hi <= 10), show distribution
    if lo >= 0 and hi <= 10:
        uniq, counts = np.unique(arr, return_counts=True)
        weights = counts.astype('d') / arr.size

        if range_ == uniq.size:  # 100% range, no skips
            yield f'{dtype}∈[{lo} ... {hi}] @ {_fmt_1d(weights)}'
        else:
            yield f'{dtype}∈{uniq} @ {_fmt_1d(weights)}'

    else:  # Wide range - low/high + mean/std + nuniq (opt) + gradient
        yield from (
            f'{dtype}({arr.mean():.3g} ± {arr.std():.3g})',
            f'X∈[{lo} ... {hi}]',
        )

        # Not much uniqs
        if (range_ < 1_000 or arr.size < 1_000_000 or arr.itemsize <= 2) and (
            nuniq := np.unique(arr).size
        ) != range_:
            yield f'{nuniq / range_:.2%} range'


def _float_info(arr: np.ndarray, lo, hi, dtype) -> Iterator[str]:
    if dtype.kind == 'c':  # Force complex as float
        arr = arr.astype('F').view('2f')
        lo, hi = arr.min(), arr.max()  # Complex min/max uses amplitude

    arr = np.ma.masked_invalid(arr)
    mask = np.ma.getmask(arr)
    if (num_invalid := mask.sum()) < arr.size:
        if num_invalid:  # Old min/max have invalid data, recompute
            lo, hi = arr.min(), arr.max()
            yield f'{num_invalid / arr.size:.2%} invalid'

        if lo < hi:
            yield from (
                f'{dtype}({arr.mean():.3g} ± {arr.std():.3g})',
                f'X∈[{lo:.3g} ... {hi:.3g}]',
            )
        else:
            yield f'{dtype}({lo:.3g})'

    # NaN/-Inf/+Inf
    yield from map(str, np.unique(arr.data[mask]).tolist())


def _grad_info(arr: np.ndarray | np.ma.MaskedArray) -> Iterator[str]:
    if (grad := _get_nd_grad(arr)).any():
        yield f'grad={_fmt_1d(grad)}'


def _get_properties(arr: np.ndarray, lo, hi) -> Iterator[str]:
    dtype: np.dtype = arr.dtype
    assert arr.size
    match dtype.kind:
        case 'b':  # Bool
            yield from _bool_info(arr)

        case 'u' | 'i':  # Integers
            yield from _int_info(arr, lo, hi, dtype)
            yield from _grad_info(arr)

        case 'c' | 'f':  # Dense data, use mean/std/gradient
            yield from _float_info(arr, lo, hi, dtype)
            yield from _grad_info(arr)

        case _:
            msg = f'Unknown dtype: {dtype}'
            raise NotImplementedError(msg)


class _ReprArray(NamedTuple):
    data: np.ndarray

    def __str__(self) -> str:
        return str(self.data)

    def __repr__(self) -> str:
        arr = self.data
        if not arr.size:
            return f'{arr!r}'  # ndarray.__repr__ is short for 0-sized arrays
        dtype = arr.dtype

        # Single value
        if arr.size == 1:
            return f'np.{dtype}({arr})'

        # Try to collapse
        lo = hi = None
        shape = f'{arr.size if arr.ndim == 1 else arr.shape}'
        if dtype.kind in 'buifc':
            lo, hi = arr.min(), arr.max()
            if np.isfinite([lo, hi]).all() and lo == hi:  # "full" array
                match lo:
                    case 0:
                        template = 'np.zeros({shape}, {dtype})'
                    case 1:
                        template = 'np.ones({shape}, {dtype})'
                    case _:
                        template = 'np.full({shape}, {dtype}({value}))'
                return template.format(shape=shape, dtype=dtype, value=lo)

        # Small array (or unknown dtype), print contents as is
        if arr.size < _MAX_SIZES.get(dtype.kind, np.inf):
            if arr.ndim == 1 or arr.size == 1:
                return f'np.{dtype}({arr})'
            return f'np.array({shape}, {dtype}({arr.ravel()}))'

        return (
            'np.array('
            + ', '.join([shape, *_get_properties(arr, lo, hi)])
            + ')'
        )


def _prepare(obj):  # noqa: PLR0911
    if isinstance(obj, np.ndarray):
        return _ReprArray(obj)

    if isinstance(obj, str | bytes | bytearray | range | Iterator):
        return obj

    if is_dataclass(obj):
        items = {f.name: getattr(obj, f.name) for f in fields(obj) if f.init}
        return replace(obj, **_prepare(items))  # type: ignore

    # namedtuple
    if isinstance(obj, tuple) and hasattr(obj, '_fields'):
        return type(obj)(*map(_prepare, obj))

    if isinstance(obj, Mapping):
        return {_prepare(k): _prepare(v) for k, v in obj.items()}

    if isinstance(obj, Iterable):
        return type(obj)(map(_prepare, obj))  # type: ignore[call-arg]

    return obj


def ic_repr(obj: Any, width: int | None = None) -> str:
    obj_repr = _prepare(obj)

    if width is None:
        width = shutil.get_terminal_size().columns
    line = pprint.pformat(obj_repr, width=width)

    # Preserve string newlines in output.
    return line.replace(r'\n', '\n')


def _format_time() -> str:
    now = f'{datetime.now():%H:%M:%S.%f}'[:-3]  # Tail is not usecs, but msecs
    return f' at {now}'


def _format_context(frame: FrameType, call_node: EnhancedAST | None) -> str:
    info = inspect.getframeinfo(frame)
    parent_fn = info.function

    if parent_fn != '<module>':
        parent_fn = f'{parent_fn}()'

    if call_node is None:
        return f'{basename(info.filename)} in {parent_fn}'

    lineno = call_node.lineno  # type: ignore[attr-defined]
    return f'{basename(info.filename)}:{lineno} in {parent_fn}'


def _construct_argument_output(
    context: str, pairs: Iterable[tuple[str, Any]]
) -> str:
    pairs = [(arg, ic_repr(val)) for arg, val in pairs]
    # For cleaner output, if <arg> is a literal, eg 3, "string", b'bytes',
    # etc, only output the value, not the argument and the value, as the
    # argument and the value will be identical or nigh identical. Ex: with
    # ic("hello"), just output
    #
    #   ic| 'hello',
    #
    # instead of
    #
    #   ic| "hello": 'hello'.
    #
    single_line_formatted_args = ', '.join(
        val if is_literal(arg) else f'{arg}: {val}' for arg, val in pairs
    )

    if len(single_line_formatted_args.splitlines()) <= 1:
        all_pairs = (
            f'{PREFIX}{context} - {single_line_formatted_args}'
            if context
            else f'{PREFIX}{single_line_formatted_args}'
        )
        if len(all_pairs.splitlines()[0]) <= LINE_WRAP_WIDTH:
            # ic| foo.py:11 in foo() - a: 1, b: 2
            # ic| a: 1, b: 2, c: 3
            return all_pairs

    lines = tuple(format_pair(arg, value) for arg, value in pairs)
    if context:
        # ic| foo.py:11 in foo()
        #     multilineStr: 'line1
        #                    line2'
        #
        # ic| foo.py:11 in foo()
        #     a: 11111111111111111111
        #     b: 22222222222222222222
        lines = context, *lines
    else:
        # ic| multilineStr: 'line1
        #                    line2'
        #
        # ic| a: 11111111111111111111
        #     b: 22222222222222222222
        pass
    return '\n'.join(indented_lines(PREFIX, '\n'.join(lines)))


def _format(frame: FrameType, *args) -> str:
    call_node = Source.executing(frame).node
    context = _format_context(frame, call_node)
    if not args:
        return PREFIX + context + _format_time()

    pairs: Iterable[tuple[str, Any]]
    if call_node is None:
        pairs = ((f'{i}', arg) for i, arg in enumerate(args))
    else:
        source: Source = Source.for_frame(frame)  # type: ignore[assignment]
        sanitized_arg_strs = [
            source.get_text_with_indentation(arg)
            for arg in call_node.args  # type: ignore[attr-defined]
        ]
        pairs = zip(sanitized_arg_strs, args)

    return _construct_argument_output(context, pairs)


@overload
def ic() -> None: ...


@overload
def ic[T](x: T, /) -> T: ...


@overload
def ic[T, T2, *Ts](x: T, x2: T2, /, *xs: *Ts) -> tuple[T, T2, *Ts]: ...


def ic(*args):
    frame = inspect.currentframe()
    assert frame
    assert frame.f_back
    out = _format(frame.f_back, *args)

    s = highlight(out, LEXER, FORMATTER)
    with LOCK:
        print(s, file=sys.stderr)

    if not args:
        return None  # E.g. ic().
    if len(args) == 1:
        return args[0]  # E.g. ic(1).
    return args  # E.g. ic(1, 2, 3).
